# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import falcon
from monasca_common.kafka import producer
from monasca_common.kafka_lib.common import FailedPayloadsError
from monasca_events_api.app.model import envelope as ev_envelope
from monasca_events_api import conf
from oslo_log import log
from oslo_utils import encodeutils


LOG = log.getLogger(__name__)
CONF = conf.CONF

_RETRY_AFTER = 60
_KAFKA_META_DATA_SIZE = 32
_TRUNCATION_SAFE_OFFSET = 1


class EventPublisher(object):
    """Publishes events data to Kafka

    EventPublisher is able to send single message to multiple configured topic.
    It uses following configuration written in conf file ::

        [event_publisher]
        topics = 'monevents'
        kafka_url = 'localhost:8900'

    Note:
        Uses :py:class:`monasca_common.kafka.producer.KafkaProducer`
        to ship events to kafka. For more details
        see `monasca-common`_ github repository.

    .. _monasca-common: https://github.com/openstack/monasca-common

    """

    def __init__(self):

        self._topics = CONF.events_publisher.topics

        self._kafka_publisher = producer.KafkaProducer(
            url=CONF.events_publisher.kafka_url
        )

        LOG.info('Initializing EventPublisher <%s>', self)

    def _transform_message(self, message):
        """Serialize and ensure that message has proper type

        :param str message: instance of message
        :return: serialized message
        :rtype: str
        """
        serialized = ev_envelope.serialize_envelope(message)
        return encodeutils.safe_encode(serialized, 'utf-8')

    def _publish(self, messages):
        """Publishes messages to kafka.

        :param list messages: list of messages
        """
        num_of_msg = len(messages)

        LOG.debug('Publishing %d messages', num_of_msg)

        first = True
        while True:
            try:
                for topic in self._topics:
                    self._kafka_publisher.publish(
                        topic,
                        messages
                    )
                    LOG.debug('Sent %d messages to topic %s',
                              num_of_msg, topic)
                break
            except FailedPayloadsError as ex:
                # FailedPayloadsError exception can be cause by connection
                # problem, to make sure that is not connection issue
                # message is sent again.
                LOG.error('Failed to send messages %s', ex)
                if first:
                    LOG.error('Retrying')
                    first = False
                    continue
                else:
                    raise falcon.HTTPServiceUnavailable('Service unavailable',
                                                        str(ex), 60)
            except Exception as ex:
                LOG.error('Failed to send messages %s', ex)
                raise falcon.HTTPServiceUnavailable('Service unavailable',
                                                    str(ex), 60)

    @staticmethod
    def _is_message_valid(message):
        """Validates message before sending.

        Methods checks if message is :py:class:`model.envelope.Envelope`.
        By being instance of this class it is ensured that all required
        keys are found and they will have their values.

        """
        return isinstance(message, ev_envelope.Envelope)

    def _check_if_all_messages_was_publish(self, send_count, to_send_count):
        """Executed after publishing to sent metrics.

        :param int send_count: how many messages have been sent
        :param int to_send_count: how many messages should be sent

        """
        failed_to_send = to_send_count - send_count

        if failed_to_send == 0:
            LOG.debug('Successfully published all [%d] messages',
                      send_count)
        else:
            error_str = ('Failed to send all messages, %d '
                         'messages out of %d have not been published')
            LOG.error(error_str, failed_to_send, to_send_count)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Use this file for deploying the API under mod_wsgi.
"""

from paste import deploy

application = None


def main():
    base_dir = '/etc/monasca/'
    conf = '{0}events-api-paste.ini'.format(base_dir)
    app = deploy.loadapp('config:{0}'.format(conf))
    return app

if __name__ == '__main__' or __name__.startswith('_mod_wsgi'):
    application = main()



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import pkgutil

from oslo_log import log
from oslo_utils import importutils

LOG = log.getLogger(__name__)
_BASE_MOD_PATH = 'monasca_events_api.policies.'


def load_policy_modules():
    """Load all modules that contain policies.

    Method iterates over modules of :py:mod:`monasca_events_api.policies`
    and imports only those that contain following methods:

    - list_rules

    """
    for modname in _list_module_names():
        mod = importutils.import_module(_BASE_MOD_PATH + modname)
        if hasattr(mod, 'list_rules'):
            yield mod


def _list_module_names():
    package_path = os.path.dirname(os.path.abspath(__file__))
    for _, modname, ispkg in pkgutil.iter_modules(path=[package_path]):
        if not (modname == "opts" and ispkg):
            yield modname


def list_rules():
    """List all policy modules rules.

    Goes through all policy modules and yields their rules

    """
    all_rules = []
    for mod in load_policy_modules():
        rules = mod.list_rules()
        all_rules.extend(rules)
    return all_rules



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from monasca_common.policy import policy_engine as policy
from oslo_context import context

from monasca_events_api import policies

policy.POLICIES = policies


class RequestContext(context.RequestContext):
    """RequestContext.

    RequestContext is customized version of
    :py:class:oslo_context.context.RequestContext.
    """

    def to_policy_values(self):
        pl = super(RequestContext, self).to_policy_values()
        pl['is_admin'] = self.is_admin
        return pl

    def can(self, action, target=None):

        if target is None:
            target = {'project_id': self.project_id,
                      'user_id': self.user_id}

        return policy.authorize(self, action=action, target=target)



# Copyright 2015 kornicameister@gmail.com
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import falcon
import six

from monasca_common.rest import utils as rest_utils


_VERSIONS_TPL_DICT = {
    'v1.0': {
        'id': 'v1.0',
        'links': [
            {
                'rel': 'event',
                'href': '/events'
            }
        ],
        'status': 'CURRENT',
        'updated': "2017-09-01T00:00:00Z"
    },
}


class Versions(object):
    """Versions Api"""

    @staticmethod
    def handle_none_version_id(req, res, result):
        for version in _VERSIONS_TPL_DICT:
            selected_version = _parse_version(version, req)
            result['elements'].append(selected_version)
        res.body = rest_utils.as_json(result, sort_keys=True)
        res.status = falcon.HTTP_200

    @staticmethod
    def handle_version_id(req, res, result, version_id):
        if version_id in _VERSIONS_TPL_DICT:
            result['elements'].append(_parse_version(version_id, req))
            res.body = rest_utils.as_json(result, sort_keys=True)
            res.status = falcon.HTTP_200
        else:
            error_body = {'message': '%s is not valid version' % version_id}
            res.body = rest_utils.as_json(error_body)
            res.status = falcon.HTTP_400

    def on_get(self, req, res, version_id=None):
        result = {
            'links': _get_common_links(req),
            'elements': []
        }
        if version_id is None:
            self.handle_none_version_id(req, res, result)
        else:
            self.handle_version_id(req, res, result, version_id)


def _get_common_links(req):
    self_uri = req.uri
    if six.PY2:
        self_uri = self_uri.decode(rest_utils.ENCODING)
    base_uri = self_uri.replace(req.path, '')
    return [
        {
            'rel': 'self',
            'href': self_uri
        },
        {
            'rel': 'version',
            'href': '%s/version' % base_uri
        },
        {
            'rel': 'healthcheck',
            'href': '%s/healthcheck' % base_uri
        }
    ]


def _parse_version(version_id, req):
    self_uri = req.uri
    if six.PY2:
        self_uri = self_uri.decode(rest_utils.ENCODING)
    base_uri = self_uri.replace(req.path, '')

    # need to get template dict, consecutive calls
    # needs to operate on unmodified instance

    selected_version = _VERSIONS_TPL_DICT[version_id].copy()
    raw_links = selected_version['links']
    links = []

    for link in raw_links:
        raw_link_href = link.get('href')
        raw_link_rel = link.get('rel')
        link_href = base_uri + '/' + version_id + raw_link_href
        links.append({
            'href': link_href,
            'rel': raw_link_rel
        })
    selected_version['links'] = links

    return selected_version



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

import falcon
from monasca_common.rest import utils as rest_utils

from monasca_events_api.app.healthcheck import kafka_check

HealthCheckResult = collections.namedtuple('HealthCheckResult',
                                           ['status', 'details'])


class HealthChecks(object):
    # response configuration
    CACHE_CONTROL = ['must-revalidate', 'no-cache', 'no-store']

    # response codes
    HEALTHY_CODE_GET = falcon.HTTP_OK
    HEALTHY_CODE_HEAD = falcon.HTTP_NO_CONTENT
    NOT_HEALTHY_CODE = falcon.HTTP_SERVICE_UNAVAILABLE

    def __init__(self):
        self._kafka_check = kafka_check.KafkaHealthCheck()
        super(HealthChecks, self).__init__()

    def on_head(self, req, res):
        res.status = self.HEALTHY_CODE_HEAD
        res.cache_control = self.CACHE_CONTROL

    def on_get(self, req, res):
        # at this point we know API is alive, so
        # keep up good work and verify kafka status

        kafka_result = self._kafka_check.healthcheck()

        # in case it'd be unhealthy,
        # message will contain error string
        status_data = {
            'kafka': kafka_result.message
        }

        # Really simple approach, ideally that should be
        # part of monasca-common with some sort of registration of
        # healthchecks concept

        res.status = (self.HEALTHY_CODE_GET
                      if kafka_result.healthy else self.NOT_HEALTHY_CODE)
        res.cache_control = self.CACHE_CONTROL
        res.body = rest_utils.as_json(status_data)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from falcon import testing

from monasca_events_api.app.core import request
from monasca_events_api.tests.unit import base


class TestRequest(base.BaseTestCase):

    def setUp(self):
        super(TestRequest, self).setUp()

    def test_use_context_from_request(self):
        req = request.Request(
            testing.create_environ(
                path='/',
                headers={
                    'X_AUTH_TOKEN': '1111',
                    'X_USER_ID': '2222',
                    'X_PROJECT_ID': '3333',
                    'X_ROLES': 'goku,vegeta'
                }
            )
        )
        self.assertEqual('1111', req.context.auth_token)
        self.assertEqual('2222', req.context.user_id)
        self.assertEqual('3333', req.context.project_id)
        self.assertItemsEqual(['goku', 'vegeta'], req.context.roles)

    def test_check_is_admin_from_request(self):
        req = request.Request(
            testing.create_environ(
                path='/',
                headers={
                    'X_USER_ID': '2222',
                    'X_PROJECT_ID': '3333',
                    'X_ROLES': 'admin,burger'
                }
            ),

        )
        self.assertTrue(req.is_admin)

    def test_request_context_admin_uppercase(self):
        req = request.Request(
            testing.create_environ(
                path='/',
                headers={
                    'X_USER_ID': '1111',
                    'X_PROJECT_ID': '2222',
                    'X_ROLES': 'Admin,bob'
                }
            )
        )
        self.assertTrue(req.is_admin)



# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from monasca_common.rest import utils as rest_utils
from monasca_events_api.app.common import events_publisher
from monasca_events_api.app.model import envelope
from monasca_events_api import conf
from oslo_log import log
from oslo_utils import encodeutils

LOG = log.getLogger(__name__)
CONF = conf.CONF


class EventsBulkProcessor(events_publisher.EventPublisher):
    """BulkProcessor for effective events processing and publishing.

    BulkProcessor is customized version of
    :py:class:`monasca_events_api.app.base.event_publisher.EventPublisher`
    that utilizes processing of bulk request inside single loop.

    """

    def send_message(self, events, event_project_id=None):
        """Sends bulk package to kafka

        :param list events: received events
        :param str event_project_id: project id
        """

        num_of_msgs = len(events) if events else 0
        to_send_msgs = []

        LOG.debug('Bulk package <events=%d>',
                  num_of_msgs)

        for ev_el in events:
            try:
                t_el = self._transform_message(ev_el, event_project_id)
                if t_el:
                    to_send_msgs.append(t_el)
            except Exception as ex:
                LOG.error('Failed to transform message to json. '
                          'message: {} Exception {}'.format(ev_el, str(ex)))

        sent_count = len(to_send_msgs)
        try:
            self._publish(to_send_msgs)
        except Exception as ex:
            LOG.error('Failed to send bulk package <events=%d, dimensions=%s>',
                      num_of_msgs)
            LOG.exception(ex)
            raise ex
        finally:
            self._check_if_all_messages_was_publish(num_of_msgs, sent_count)

    def _transform_message(self, event_element, event_project_id):
        """Transform the message

        :param dict event_element: original event element
        :param str event_project_id: project id
        :return: message payload
        """
        try:
            msg_json = rest_utils.as_json(event_element)
            msg_json = encodeutils.safe_encode(msg_json, 'utf-8')

            event_envelope = envelope.Envelope.new_envelope(
                event=msg_json,
                project_id=event_project_id,
            )

            msg_payload = (super(EventsBulkProcessor, self)
                           ._transform_message(event_envelope))
            return msg_payload

        except Exception as ex:
            LOG.error("Event transformation failed, rejecting event")
            LOG.exception(ex)
            return None



# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from monasca_common.rest import utils as rest_utils
from oslo_utils import encodeutils
from oslo_utils import timeutils


def serialize_envelope(envelope):
    """Returns json representation of an envelope.

    :return: json object of envelope
    :rtype: json or a bytestring `encoding` encoded
            representation of it.

    """
    json = rest_utils.as_json(envelope, ensure_ascii=False)
    return encodeutils.safe_decode(json, 'utf-8')


class EventEnvelopeException(Exception):
    pass


class Envelope(dict):
    def __init__(self, event, meta):
        if not event:
            error_msg = 'Envelope cannot be created without event'
            raise EventEnvelopeException(error_msg)
        if 'project_id' not in meta or not meta.get('project_id'):
            error_msg = 'Envelope cannot be created without project_id'
            raise EventEnvelopeException(error_msg)

        creation_time = self._get_creation_time()
        super(Envelope, self).__init__(
            event=rest_utils.from_json(event),
            creation_time=creation_time,
            meta=meta
        )

    @staticmethod
    def _get_creation_time():
        return timeutils.utcnow_ts()

    @classmethod
    def new_envelope(cls, event, project_id):
        """Creates new event envelope

        Event envelope is combined of of following properties

        * event - dict
        * creation_time - timestamp
        * meta - meta block

        Example output json would be like this:

        .. code-block:: json

            {
                "event": {
                  "message": "Some message",
                  "dimensions": {
                    "hostname": "devstack"
                  }
                },
                "creation_time": 1447834886,
                "meta": {
                  "project_id": "e4bd29509eda473092d32aadfee3e7b1",
                }
            }

        :param dict event: original event element
        :param str project_id: project id to be put in meta field
        """
        event_meta = {
            'project_id': project_id
        }

        return cls(event, event_meta)

    @property
    def event(self):
        return self.get('event', None)

    @property
    def creation_time(self):
        return self.get('creation_time', None)

    @property
    def meta(self):
        return self.get('meta', None)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from falcon import testing

from monasca_common.policy import policy_engine as policy
from oslo_context import context
from oslo_policy import policy as os_policy

from monasca_events_api.app.core import request
from monasca_events_api.tests.unit import base


class TestPolicyFileCase(base.BaseTestCase):
    def setUp(self):
        super(TestPolicyFileCase, self).setUp()
        self.context = context.RequestContext(user='fake',
                                              tenant='fake',
                                              is_admin=False)
        self.target = {'tenant_id': 'fake'}

    def test_modified_policy_reloads(self):
        tmp_file = \
            self.create_tempfiles(files=[('policies', '{}')], ext='.yaml')[0]
        base.BaseTestCase.conf_override(policy_file=tmp_file,
                                        group='oslo_policy')

        policy.reset()
        policy.init()

        action = 'example:test'
        rule = os_policy.RuleDefault(action, '')
        policy._ENFORCER.register_defaults([rule])

        with open(tmp_file, 'w') as policy_file:
            policy_file.write('{"example:test": ""}')
        policy.authorize(self.context, action, self.target)

        with open(tmp_file, 'w') as policy_file:
            policy_file.write('{"example:test": "!"}')
        policy._ENFORCER.load_rules(True)
        self.assertRaises(os_policy.PolicyNotAuthorized, policy.authorize,
                          self.context, action, self.target)


class TestPolicyCase(base.BaseTestCase):
    def setUp(self):
        super(TestPolicyCase, self).setUp()
        rules = [
            os_policy.RuleDefault("true", "@"),
            os_policy.RuleDefault("example:allowed", "@"),
            os_policy.RuleDefault("example:denied", "!"),
            os_policy.RuleDefault("example:lowercase_admin",
                                  "role:admin or role:sysadmin"),
            os_policy.RuleDefault("example:uppercase_admin",
                                  "role:ADMIN or role:sysadmin"),
        ]
        policy.reset()
        policy.init()
        policy._ENFORCER.register_defaults(rules)

    def test_authorize_nonexist_action_throws(self):
        action = "example:noexist"
        ctx = request.Request(
            testing.create_environ(
                path="/",
                headers={
                    "X_USER_ID": "fake",
                    "X_PROJECT_ID": "fake",
                    "X_ROLES": "member"
                }
            )
        )
        self.assertRaises(os_policy.PolicyNotRegistered, policy.authorize,
                          ctx.context, action, {})

    def test_authorize_bad_action_throws(self):
        action = "example:denied"
        ctx = request.Request(
            testing.create_environ(
                path="/",
                headers={
                    "X_USER_ID": "fake",
                    "X_PROJECT_ID": "fake",
                    "X_ROLES": "member"
                }
            )
        )
        self.assertRaises(os_policy.PolicyNotAuthorized, policy.authorize,
                          ctx.context, action, {})

    def test_authorize_bad_action_no_exception(self):
        action = "example:denied"
        ctx = request.Request(
            testing.create_environ(
                path="/",
                headers={
                    "X_USER_ID": "fake",
                    "X_PROJECT_ID": "fake",
                    "X_ROLES": "member"
                }
            )
        )
        result = policy.authorize(ctx.context, action, {}, False)
        self.assertFalse(result)

    def test_authorize_good_action(self):
        action = "example:allowed"
        ctx = request.Request(
            testing.create_environ(
                path="/",
                headers={
                    "X_USER_ID": "fake",
                    "X_PROJECT_ID": "fake",
                    "X_ROLES": "member"
                }
            )
        )
        result = policy.authorize(ctx.context, action, False)
        self.assertTrue(result)

    def test_ignore_case_role_check(self):
        lowercase_action = "example:lowercase_admin"
        uppercase_action = "example:uppercase_admin"

        admin_context = request.Request(
            testing.create_environ(
                path="/",
                headers={
                    "X_USER_ID": "admin",
                    "X_PROJECT_ID": "fake",
                    "X_ROLES": "AdMiN"
                }
            )
        )
        self.assertTrue(policy.authorize(admin_context.context,
                                         lowercase_action,
                                         {}))
        self.assertTrue(policy.authorize(admin_context.context,
                                         uppercase_action,
                                         {}))



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from six.moves.urllib.parse import urlparse as urlparse

import falcon

from monasca_events_api.app.controller import versions
from monasca_events_api.tests.unit import base


def _get_versioned_url(version_id):
    return '/version/%s' % version_id


class TestVersionApi(base.BaseApiTestCase):

    def setUp(self):
        super(TestVersionApi, self).setUp()
        self.versions = versions.Versions()
        self.app.add_route("/version/", self.versions)
        self.app.add_route("/version/{version_id}", self.versions)

    def test_request_for_incorrect_version(self):
        incorrect_version = 'v2.0'
        uri = _get_versioned_url(incorrect_version)

        res = self.simulate_request(
            path=uri,
            method='GET',
            headers={
                'Content-Type': 'application/json'
            }
        )

        self.assertEqual(falcon.HTTP_400, res.status)

    def test_should_return_supported_event_api_version(self):

        def _check_global_links(current_endpoint, links):
            expected_links = {'self': current_endpoint,
                              'version': '/version',
                              'healthcheck': '/healthcheck'}
            _check_links(links, expected_links)

        def _check_links(links, expected_links):
            for link in links:
                self.assertIn('rel', link)
                self.assertIn('href', link)
                key = link.get('rel')
                href_path = urlparse(link.get('href')).path
                self.assertIn(key, expected_links.keys())
                self.assertEqual(expected_links[key], href_path)

        def _check_elements(elements, expected_versions):
            self.assertIsInstance(elements, list)
            for el in elements:
                self.assertItemsEqual([
                    u'id',
                    u'links',
                    u'status',
                    u'updated'
                ], el.keys())
                id_v = el.get('id')
                self.assertEqual(expected_versions, id_v)

        supported_versions = ['v1.0']
        version_endpoint = '/version'
        for version in supported_versions:
            endpoint = '%s/%s' % (version_endpoint, version)
            res = self.simulate_request(
                path=endpoint,
                method='GET',
                headers={
                    'Content-Type': 'application/json'
                }
            )
            self.assertEqual(falcon.HTTP_200, res.status)
            response = res.json
            self.assertIn('links', response)
            _check_global_links(endpoint, response['links'])
            self.assertIn('elements', response)
            _check_elements(response['elements'], version)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import falcon
import mock

from monasca_events_api.app.controller import healthchecks
from monasca_events_api.app.healthcheck import kafka_check as healthcheck
from monasca_events_api.tests.unit import base

ENDPOINT = '/healthcheck'


class TestApiHealthChecks(base.BaseApiTestCase):

    def setUp(self):
        super(TestApiHealthChecks, self).setUp()
        self.resource = healthchecks.HealthChecks()
        self.app.add_route(
            ENDPOINT,
            self.resource
        )

    def test_should_return_200_for_head(self):
        res = self.simulate_request(
            path=ENDPOINT,
            method='HEAD')
        self.assertEqual(falcon.HTTP_NO_CONTENT, res.status)

    @mock.patch('monasca_events_api.app.healthcheck.'
                'kafka_check.KafkaHealthCheck')
    def test_should_report_healthy_if_kafka_healthy(self, kafka_check):
        kafka_check.healthcheck.return_value = healthcheck.CheckResult(True,
                                                                       'OK')
        self.resource._kafka_check = kafka_check

        res = self.simulate_request(path=ENDPOINT,
                                    headers={
                                        'Content-Type': 'application/json'
                                    },
                                    method='GET')
        self.assertEqual(falcon.HTTP_OK, res.status)

        res = res.json
        self.assertIn('kafka', res)
        self.assertEqual('OK', res.get('kafka'))

    @mock.patch('monasca_events_api.app.healthcheck.'
                'kafka_check.KafkaHealthCheck')
    def test_should_report_unhealthy_if_kafka_unhealthy(self, kafka_check):
        url = 'localhost:8200'
        err_str = 'Could not connect to kafka at %s' % url
        kafka_check.healthcheck.return_value = healthcheck.CheckResult(False,
                                                                       err_str)
        self.resource._kafka_check = kafka_check

        res = self.simulate_request(path=ENDPOINT,
                                    headers={
                                        'Content-Type': 'application/json'
                                    },
                                    method='GET')
        self.assertEqual(falcon.HTTP_SERVICE_UNAVAILABLE, res.status)

        res = res.json
        self.assertIn('kafka', res)
        self.assertEqual(err_str, res.get('kafka'))



# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import falcon
from monasca_common.policy import policy_engine as policy
from oslo_log import log

from monasca_events_api.app.core import request_contex
from monasca_events_api.app.core import validation
from monasca_events_api import policies

LOG = log.getLogger(__name__)
policy.POLICIES = policies


class Request(falcon.Request):
    """Variation of falcon. Request with context.

    Following class enhances :py:class:`falcon.Request` with
    :py:class:`context.CustomRequestContext`
    """

    def __init__(self, env, options=None):
        """Init an Request class."""
        super(Request, self).__init__(env, options)
        self.context = \
            request_contex.RequestContext.from_environ(self.env)
        self.is_admin = policy.check_is_admin(self.context)
        self.project_id = self.context.project_id

    def validate(self, content_types):
        """Performs common request validation

        Validation checklist (in that order):

        * :py:func:`validation.validate_content_type`

        :param content_types: allowed content-types handler supports
        :type content_types: list
        :raises Exception: if any of the validation fails

        """
        validation.validate_content_type(self, content_types)

    def can(self, action, target=None):
        return self.context.can(action, target)



# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import falcon

from monasca_events_api.app.common import helpers
from monasca_events_api.app.controller.v1 import body_validation
from monasca_events_api.app.controller.v1 import bulk_processor
from monasca_events_api.app.core.model import prepare_message_to_sent
from oslo_log import log

LOG = log.getLogger(__name__)


class Events(object):
    """Events.

    Events acts as a RESTful endpoint accepting messages contains
    collected events from the OpenStack message bus.
    Works as getaway for any further processing for accepted data.
    """
    VERSION = 'v1.0'
    SUPPORTED_CONTENT_TYPES = {'application/json'}

    def __init__(self):
        super(Events, self).__init__()

        self._processor = bulk_processor.EventsBulkProcessor()

    def on_post(self, req, res):
        """Accepts sent events as json.

        Accepts events sent to resource which should be sent
        to Kafka queue.

        :param req: current request
        :param res: current response
        """
        policy_action = 'events_api:agent_required'

        try:
            req.validate(self.SUPPORTED_CONTENT_TYPES)
            request_body = helpers.read_json_msg_body(req)
            req.can(policy_action)
            project_id = req.project_id
            body_validation.validate_body(request_body)
            messages = prepare_message_to_sent(request_body)
            self._processor.send_message(messages, event_project_id=project_id)
            res.status = falcon.HTTP_200
        except falcon.HTTPUnprocessableEntity as ex:
            LOG.error('Entire bulk package was rejected, unsupported body')
            LOG.exception(ex)
            raise ex
        except falcon.HTTPUnsupportedMediaType as ex:
            LOG.error('Entire bulk package was rejected, '
                      'unsupported media type')
            LOG.exception(ex)
            raise ex
        except Exception as ex:
            LOG.error('Entire bulk package was rejected')
            LOG.exception(ex)
            _title = ex.title if hasattr(ex, 'title') else None
            _descr = ex.description if hasattr(ex, 'description') else None
            raise falcon.HTTPError(falcon.HTTP_400,
                                   title=_title, description=_descr)

    @property
    def version(self):
        return getattr(self, 'VERSION')



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import falcon

from oslo_log import log

from monasca_common.rest import exceptions


LOG = log.getLogger(__name__)


def read_json_msg_body(req):
    """Read the json_msg from the http request body and return as JSON.

    :param req: HTTP request object.
    :return: Returns the metrics as a JSON object.
    :raises falcon.HTTPBadRequest:
    """
    try:
        return req.media
    except exceptions.DataConversionException as ex:
        LOG.debug(ex)
        raise falcon.HTTPBadRequest('Bad request',
                                    'Request body is not valid JSON')
    except ValueError as ex:
        LOG.debug(ex)
        raise falcon.HTTPBadRequest('Bad request',
                                    'Request body is not valid JSON')



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import falcon
from falcon import testing
import fixtures
from monasca_common.policy import policy_engine as policy
from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_context import fixture as oc_fixture
from oslo_log.fixture import logging_error as log_fixture
from oslo_serialization import jsonutils
from oslotest import base

from monasca_events_api.app.core import request
from monasca_events_api import config
from monasca_events_api import policies


CONF = cfg.CONF
policy.POLICIES = policies


class ConfigFixture(config_fixture.Config):

    def setUp(self):
        super(ConfigFixture, self).setUp()
        self.addCleanup(self._clean_config_loaded_flag)
        config.parse_args()

    @staticmethod
    def _clean_config_loaded_flag():
        config._CONF_LOADED = False


class BaseTestCase(base.BaseTestCase):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.useFixture(fixtures.NestedTempfile())
        self.useFixture(fixtures.TempHomeDir())
        self.useFixture(log_fixture.get_logging_handle_error_fixture())
        self.useFixture(ConfigFixture(CONF))
        self.useFixture(oc_fixture.ClearRequestContext())
        self.useFixture(PolicyFixture())

    @staticmethod
    def conf_override(**kw):
        """Override flag variables for a test."""
        group = kw.pop('group', None)
        for k, v in kw.items():
            CONF.set_override(k, v, group)


class PolicyFixture(fixtures.Fixture):
    """Override the policy with a completely new policy file.

    This overrides the policy with a completely fake and synthetic
    policy file.

    """

    def setUp(self):
        super(PolicyFixture, self).setUp()
        self._prepare_policy()
        policy.reset()
        policy.init()
        self.addCleanup(policy.reset)

    def _prepare_policy(self):
        policy_dir = self.useFixture(fixtures.TempDir())
        policy_file = os.path.join(policy_dir.path, 'policy.yaml')

        # load the fake_policy data and add the missing default rules.
        policy_rules = jsonutils.loads('{}')
        self.add_missing_default_rules(policy_rules)
        with open(policy_file, 'w') as f:
            jsonutils.dump(policy_rules, f)

        BaseTestCase.conf_override(policy_file=policy_file,
                                   group='oslo_policy')
        BaseTestCase.conf_override(policy_dirs=[], group='oslo_policy')

    def add_missing_default_rules(self, rules):
        for rule in policies.list_rules():
            if rule.name not in rules:
                rules[rule.name] = rule.check_str


class BaseApiTestCase(BaseTestCase, testing.TestCase):

    def setUp(self):
        super(BaseApiTestCase, self).setUp()
        self.app = falcon.API(request_type=request.Request)
        # NOTE: Falcon 2.0.0 switches the default for this from
        # True to False so we explicitly set it here to prevent the behaviour
        # changing between versions.
        self.app.req_options.strip_url_path_trailing_slash = True



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import pkgutil

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils

LOG = log.getLogger(__name__)
CONF = cfg.CONF


def load_conf_modules():
    """Load all modules that contain configuration.

    Method iterates over modules of :py:mod:`monasca_events_api.conf`
    and imports only those that contain following methods:

    - list_opts (required by oslo_config.genconfig)
    - register_opts (required by :py:currentmodule:)

    """
    imported_modules = []
    for modname in _list_module_names():
        mod = importutils.import_module('monasca_events_api.conf.' + modname)
        required_funcs = ['register_opts', 'list_opts']
        for func in required_funcs:
            if not hasattr(mod, func):
                msg = ("The module 'monasca_events_api.conf.%s' should have a"
                       " '%s' function which returns"
                       " the config options."
                       % (modname, func))
                LOG.warning(msg)
            else:
                imported_modules.append(mod)

    LOG.debug('Found %d modules that contain configuration',
              len(imported_modules))

    return imported_modules


def _list_module_names():
    module_names = []
    package_path = os.path.dirname(os.path.abspath(__file__))
    for _, modname, ispkg in pkgutil.iter_modules(path=[package_path]):
        if not (modname == "opts" and ispkg):
            module_names.append(modname)
    return module_names


def register_opts(conf):
    """Register all conf modules opts.

    This method allows different modules to register
    opts according to their needs.

    """
    for mod in load_conf_modules():
        mod.register_opts(conf)


def list_opts():
    """List all conf modules opts.

    Goes through all conf modules and yields their opts

    """
    for mod in load_conf_modules():
        mod_opts = mod.list_opts()
        yield mod_opts[0], mod_opts[1]



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

_MAX_MESSAGE_SIZE = 1048576

events_publisher_opts = [
    cfg.StrOpt('kafka_url',
               required=True,
               help='Url to kafka server',
               default="127.0.0.1:9092"),
    cfg.MultiStrOpt('topics',
                    help='Consumer topics',
                    default=['monevents'],)
]

events_publisher_group = cfg.OptGroup(name='events_publisher',
                                      title='events_publisher')


def register_opts(conf):
    conf.register_group(events_publisher_group)
    conf.register_opts(events_publisher_opts, events_publisher_group)


def list_opts():
    return events_publisher_group, events_publisher_opts



# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


def prepare_message_to_sent(body):
    """prepare_message_to_sent convert message to proper format,

    :param dict body: original request body
    :return dict: prepared message for publish to kafka
    """
    timestamp = body['timestamp']
    final_body = []
    for events in body['events']:
        ev = events['event'].copy()
        ev.update({'timestamp': timestamp})
        ev.update({'dimensions': events.get('dimensions')})
        final_body.append(ev)
    return final_body



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six

from falcon import HTTPUnprocessableEntity
from oslo_log import log
from voluptuous import All
from voluptuous import Any
from voluptuous import Length
from voluptuous import MultipleInvalid
from voluptuous import Required
from voluptuous import Schema


LOG = log.getLogger(__name__)


default_schema = Schema({Required("events"): All(Any(list, dict),
                         Length(min=1)),
                         Required("timestamp"): six.text_type})


def validate_body(request_body):
    """Validate body.

     Method validate if body contain all required fields,
     and check if all value have correct type.

    :param request_body: body
    """
    try:
        default_schema(request_body)
    except MultipleInvalid as ex:
        LOG.exception(ex)
        raise HTTPUnprocessableEntity(description=ex.error_message)



# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import falcon
from oslo_log import log

LOG = log.getLogger(__name__)


def validate_content_type(req, allowed):
    """Validates content type.

    Method validates request against correct
    content type.

    If content-type cannot be established (i.e. header is missing),
    :py:class:`falcon.HTTPMissingHeader` is thrown.
    If content-type is not **application/json** or **text/plain**,
    :py:class:`falcon.HTTPUnsupportedMediaType` is thrown.


    :param falcon.Request req: current request
    :param iterable allowed: allowed content type

    :exception: :py:class:`falcon.HTTPMissingHeader`
    :exception: :py:class:`falcon.HTTPUnsupportedMediaType`
    """
    content_type = req.content_type

    LOG.debug('Content-Type is %s', content_type)

    if content_type is None or len(content_type) == 0:
        raise falcon.HTTPMissingHeader('Content-Type')

    if content_type not in allowed:
        sup_types = ', '.join(allowed)
        details = ('Only [%s] are accepted as logs representations'
                   % str(sup_types))
        raise falcon.HTTPUnsupportedMediaType(description=details)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log
from oslo_policy import opts as policy_opts

from monasca_events_api import conf
from monasca_events_api import version

CONF = cfg.CONF
LOG = log.getLogger(__name__)

_CONF_LOADED = False


def parse_args():
    """Parse configuration arguments.

    Note:
        This method ensures that configuration will be loaded
        only once within single python interpreter.

    """
    global _CONF_LOADED
    if _CONF_LOADED:
        LOG.debug('Configuration has been already loaded')
        return

    log.set_defaults()
    log.register_options(CONF)

    CONF(args=[],
         prog='events-api',
         project='monasca',
         version=version.version_str,
         description='RESTful API to collect events from cloud')

    log.setup(CONF,
              product_name='monasca-events-api',
              version=version.version_str)

    conf.register_opts(CONF)
    policy_opts.set_defaults(CONF)

    _CONF_LOADED = True



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from falcon.errors import HTTPUnprocessableEntity

from monasca_events_api.app.controller.v1.body_validation import validate_body
from monasca_events_api.tests.unit import base


class TestBodyValidation(base.BaseTestCase):

    def test_missing_events_filed(self):
        body = {'timestamp': '2012-10-29T13:42:11Z+0200'}
        self.assertRaises(HTTPUnprocessableEntity, validate_body, body)

    def test_missing_timestamp_field(self):
        body = {'events': [{'event': {'payload': 'test'}}]}
        self.assertRaises(HTTPUnprocessableEntity, validate_body, body)

    def test_empty_events_as_list(self):
        body = {'events': [], 'timestamp': u'2012-10-29T13:42:11Z+0200'}
        self.assertRaises(HTTPUnprocessableEntity, validate_body, body)

    def test_empty_events_as_dict(self):
        body = {'events': {}, 'timestamp': u'2012-10-29T13:42:11Z+0200'}
        self.assertRaises(HTTPUnprocessableEntity, validate_body, body)

    def test_empty_body(self):
        body = {}
        self.assertRaises(HTTPUnprocessableEntity, validate_body, body)

    def test_incorrect_timestamp_type(self):
        body = {'events': [], 'timestamp': 9000}
        self.assertRaises(HTTPUnprocessableEntity, validate_body, body)

    def test_incorrect_events_type(self):
        body = {'events': 'over9000', 'timestamp': '2012-10-29T13:42:11Z+0200'}
        self.assertRaises(HTTPUnprocessableEntity, validate_body, body)

    def test_correct_body(self):
        body = [{'events': [{'event': {'payload': 'test'}}],
                 'timestamp': u'2012-10-29T13:42:11Z+0200'},
                {'events': {'event': {'payload': 'test'}},
                 'timestamp': u'2012-10-29T13:42:11Z+0200'}]
        for b in body:
            validate_body(b)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Module contains factories to initializes various applications
of monasca-events-api.
"""

import falcon
from oslo_log import log
import six

from monasca_events_api.app.controller import healthchecks
from monasca_events_api.app.controller.v1 import events as v1_events
from monasca_events_api.app.controller import versions
from monasca_events_api.app.core import error_handlers
from monasca_events_api.app.core import request
from monasca_events_api import config


def error_trap(app_name):
    """Decorator trapping any error during application boot time.

    :param app_name: Application name
    :type app_name: str
    :return: _wrapper function
    """
    @six.wraps(error_trap)
    def _wrapper(func):

        @six.wraps(_wrapper)
        def _inner_wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception:
                logger = log.getLogger(__name__)
                logger.exception(
                    'Failed to load application: \'{}\''.format(app_name))
                raise
        return _inner_wrapper
    return _wrapper


def singleton_config(func):
    """Decorator ensuring that configuration is loaded only once.

    :param func: Function to execute
    :return: _wrapper
    """
    @six.wraps(singleton_config)
    def _wrapper(global_conf, **local_conf):
        config.parse_args()
        return func(global_conf, **local_conf)
    return _wrapper


@error_trap('version')
def create_version_app(global_conf, **local_conf):
    """Creates Version application"""

    ctrl = versions.Versions()
    controllers = {
        '/': ctrl,   # redirect http://host:port/ down to Version app
                     # avoid conflicts with actual pipelines and 404 error
        '/version': ctrl,  # list all the versions
        '/version/{version_id}': ctrl  # display details of the version
    }

    wsgi_app = falcon.API()
    for route, ctrl in controllers.items():
        wsgi_app.add_route(route, ctrl)
    return wsgi_app


@error_trap('healthcheck')
def create_healthcheck_app(global_conf, **local_conf):
    """Create Healthcheck application"""

    controllers = {
        '/': healthchecks.HealthChecks(),
    }

    wsgi_app = falcon.API()
    for route, ctrl in controllers.items():
        wsgi_app.add_route(route, ctrl)
    return wsgi_app


@error_trap('api')
@singleton_config
def create_api_app(global_conf, **local_conf):
    """Create Main Events Api application.

    :param global_conf: Global config
    :param local_conf: Local config
    :return: falcon.API
    """
    controllers = {}
    controllers.update({
        '/events': v1_events.Events()
    })

    wsgi_app = falcon.API(
        request_type=request.Request
    )

    for route, ctrl in controllers.items():
        wsgi_app.add_route(route, ctrl)

    error_handlers.register_error_handler(wsgi_app)

    return wsgi_app



# -*- coding: utf-8 -*-
#
# monasca-events-app documentation build configuration file, created by
# sphinx-quickstart on Wed Nov 18 12:02:03 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

sys.path = [
    os.path.abspath('../..'),
    os.path.abspath('../../bin'),
    os.path.abspath('./'),
    os.path.abspath('../')
] + sys.path

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.6'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.graphviz',
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'oslo_config.sphinxconfiggen',
    'oslo_policy.sphinxpolicygen',
    'oslo_config.sphinxext',
    'openstackdocstheme',
]

# geeneral information about project
repository_name = u'openstack/monasca-events-api'
bug_project = u'866'
bug_tag = u''
copyright = u'2017-present, OpenStack Foundation'
author = u'OpenStack Foundation'

# sample config
config_generator_config_file = [
    ('config-generator/config.conf', '_static/events-api')
]

# sample policy file
policy_generator_config_file = 'config-generator/policy.conf'
sample_policy_basename = '_static/events-api'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
source_suffix = '.rst'

# The encoding of source files.
source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = [
    'common',
    'doc',
    'documentation',
    'etc',
    'java'
]

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['monasca_events_api.', 'monasca']

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'openstackdocs'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# doc.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If false, no module index is generated.
html_use_modindex = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
#html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'MonitoringEventsApiDoc'

# -- Options for LaTeX output ---------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'MonitoringEventsApi.tex', u'Monasca Events Documentation',
     [author], 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'monitoringeventsapi', u'Monasca Events Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://doc.python.org/': None}



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

from monasca_common.kafka_lib import client
from oslo_log import log

from monasca_events_api import conf

LOG = log.getLogger(__name__)
CONF = conf.CONF


CheckResult = collections.namedtuple('CheckResult', ['healthy', 'message'])
"""Result from the healthcheck, contains healthy(boolean) and message"""


class KafkaHealthCheck(object):
    """Evaluates kafka health

    Healthcheck verifies if:

    * kafka server is up and running
    * there is a configured topic in kafka

    If following conditions are met healthcheck returns healthy status.
    Otherwise unhealthy status is returned with explanation.

     Example of middleware configuration:

    .. code-block:: ini

      [events_publisher]
      kafka_url = localhost:8900
      kafka_topics = events

    Note:
        It is possible to specify multiple topics if necessary.
        Just separate them with ,

    """

    def healthcheck(self):
        url = CONF.events_publisher.kafka_url

        try:
            kafka_client = client.KafkaClient(hosts=url)
        except client.KafkaUnavailableError as ex:
            LOG.error(repr(ex))
            error_str = 'Could not connect to kafka at %s' % url
            return CheckResult(healthy=False, message=error_str)

        result = self._verify_topics(kafka_client)
        self._disconnect_gracefully(kafka_client)

        return result

    # noinspection PyMethodMayBeStatic
    def _verify_topics(self, kafka_client):
        topics = CONF.events_publisher.topics

        for t in topics:
            # kafka client loads metadata for topics as fast
            # as possible (happens in __init__), therefore this
            # topic_partitions is sure to be filled
            for_topic = t in kafka_client.topic_partitions
            if not for_topic:
                error_str = 'Kafka: Topic %s not found' % t
                LOG.error(error_str)
                return CheckResult(healthy=False, message=error_str)

        return CheckResult(healthy=True, message='OK')

    # noinspection PyMethodMayBeStatic
    def _disconnect_gracefully(self, kafka_client):
        # at this point, client is connected so it must be closed
        # regardless of topic existence
        try:
            kafka_client.close()
        except Exception as ex:
            # log that something went wrong and move on
            LOG.error(repr(ex))



# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import falcon

from monasca_events_api.app.model import envelope


def events_envelope_exception_handlet(ex, req, resp, params):
    raise falcon.HTTPUnprocessableEntity(
        title='Failed to create Envelope',
        description=ex.message
    )


def register_error_handler(app):
    app.add_error_handler(envelope.EventEnvelopeException,
                          events_envelope_exception_handlet)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import falcon
import mock
import ujson as json

from monasca_events_api.app.controller.v1 import events
from monasca_events_api.tests.unit import base


ENDPOINT = '/events'


def _init_resource(test):
    resource = events.Events()
    test.app.add_route(ENDPOINT, resource)
    return resource


@mock.patch('monasca_events_api.app.controller.v1.'
            'bulk_processor.EventsBulkProcessor')
class TestEventsApi(base.BaseApiTestCase):

    def test_should_pass_simple_event(self, bulk_processor):
        events_resource = _init_resource(self)
        events_resource._processor = bulk_processor
        unit_test_patch = os.path.dirname(__file__)
        json_file_path = 'event_template_json/req_simple_event.json'
        patch_to_req_simple_event_file = os.path.join(unit_test_patch,
                                                      json_file_path)
        with open(patch_to_req_simple_event_file, 'r') as fi:
            body = fi.read()
        response = self.simulate_request(
            path=ENDPOINT,
            method='POST',
            headers={
                'Content-Type': 'application/json',
                'X_ROLES': 'monasca-user'
            },
            body=body
        )
        self.assertEqual(falcon.HTTP_200, response.status)

    def test_should_multiple_events(self, bulk_processor):
        events_resource = _init_resource(self)
        events_resource._processor = bulk_processor
        unit_test_patch = os.path.dirname(__file__)
        json_file_path = 'event_template_json/req_multiple_events.json'
        req_multiple_events_json = os.path.join(unit_test_patch,
                                                json_file_path)
        with open(req_multiple_events_json, 'r') as fi:
            body = fi.read()
        response = self.simulate_request(
            path=ENDPOINT,
            method='POST',
            headers={
                'Content-Type': 'application/json',
                'X_ROLES': 'monasca-user'
            },
            body=body
        )
        self.assertEqual(falcon.HTTP_200, response.status)

    def test_should_fail_empty_body(self, bulk_processor):
        events_resource = _init_resource(self)
        events_resource._processor = bulk_processor
        response = self.simulate_request(
            path=ENDPOINT,
            method='POST',
            headers={
                'Content-Type': 'application/json',
                'X_ROLES': 'monasca-user'
            },
            body=''
        )
        self.assertEqual(falcon.HTTP_422, response.status)

    def test_should_fail_missing_timestamp_in_body(self, bulk_processor):
        events_resource = _init_resource(self)
        events_resource._processor = bulk_processor
        unit_test_patch = os.path.dirname(__file__)
        json_file_path = 'event_template_json/req_simple_event.json'
        patch_to_req_simple_event_file = os.path.join(unit_test_patch,
                                                      json_file_path)
        with open(patch_to_req_simple_event_file, 'r') as fi:
            events = json.load(fi)['events']
        body = {'events': [events]}
        response = self.simulate_request(
            path=ENDPOINT,
            method='POST',
            headers={
                'Content-Type': 'application/json',
                'X_ROLES': 'monasca-user'
            },
            body=json.dumps(body)
        )
        self.assertEqual(falcon.HTTP_422, response.status)

    def test_should_fail_missing_events_in_body(self, bulk_processor):
        events_resource = _init_resource(self)
        events_resource._processor = bulk_processor
        body = {'timestamp': '2012-10-29T13:42:11Z+0200'}
        response = self.simulate_request(
            path=ENDPOINT,
            method='POST',
            headers={
                'Content-Type': 'application/json',
                'X_ROLES': 'monasca-user'
            },
            body=json.dumps(body)
        )
        self.assertEqual(falcon.HTTP_422, response.status)

    def test_should_fail_missing_content_type(self, bulk_processor):
        events_resource = _init_resource(self)
        events_resource._processor = bulk_processor
        body = {'timestamp': '2012-10-29T13:42:11Z+0200'}
        response = self.simulate_request(
            path=ENDPOINT,
            method='POST',
            headers={
                'X_ROLES': 'monasca-user'
            },
            body=json.dumps(body)
        )
        self.assertEqual(falcon.HTTP_400, response.status)

    def test_should_fail_wrong_content_type(self, bulk_processor):
        events_resource = _init_resource(self)
        events_resource._processor = bulk_processor
        body = {'timestamp': '2012-10-29T13:42:11Z+0200'}
        response = self.simulate_request(
            path=ENDPOINT,
            method='POST',
            headers={
                'Content-Type': 'text/plain',
                'X_ROLES': 'monasca-user'
            },
            body=json.dumps(body)
        )
        self.assertEqual(falcon.HTTP_415, response.status)


class TestApiEventsVersion(base.BaseApiTestCase):
    @mock.patch('monasca_events_api.app.controller.v1.'
                'bulk_processor.EventsBulkProcessor')
    def test_should_return_v1_as_version(self, _):
        resource = events.Events()
        self.assertEqual('v1.0', resource.version)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy


agent_policies = [
    policy.DocumentedRuleDefault(
        name='events_api:agent_required',
        check_str='role:monasca-user or role:admin',
        description='Send events to api',
        operations=[{'path': '/v1.0/events', 'method': 'POST'}]
    )
]


def list_rules():
    """List policies rules for agent access."""
    return agent_policies
